#!/usr/bin/perl

# APT Friend Finder
# by Pastor Manul Laphroaig
# http://www.aptfriendfinder.com/

#
#  Read file character by character (default UTF-8). 
#   For all encountered valid characters print their greedy translations
#
#  Some tables of CJK Unicode:
#  http://www.khngai.com/chinese/charmap/tbluni.php?page=5
#

#  To install Lingua::ZH::CEDICT, do "cpan Lingua::ZH::CEDICT" 

#
#  BUGS: -b gives only an approximation of the match position
#            because a decoded string forgets the original bytes
#            and is stored in Perl internal format, so translating
#            the _character_ position of match to _byte_ position
#            is hard.
#

use Lingua::ZH::CEDICT;  # contains CN dictionary (UTF-8)
use Getopt::Std;
use Encode;

use strict;

use vars qw/$opt_d $opt_f $opt_t $opt_n $opt_m $opt_b $opt_p $opt_c $opt_x/;
getopts('df:t:m:nbpcx');

my $dbg         = $opt_d;             # debug prints

# make like iconv
my $from_enc    = $opt_f || 'UTF-8';  # from encoding, see Encode::CN
my $to_enc      = $opt_t || 'UTF-8';  # to encoding,   see Encode::CN

# imitate grep options
my $line_number = $opt_n;         # print line numbers
my $max_count   = $opt_m;         # stop after printing M matching lines
my $byte_offset = $opt_b;         # print byte offsets of matches

# this one is not grep but useful
my $column_number = $opt_c;       # column number of match
my $hex           = $opt_x;       # print hex encoding of match

# what to print for matches? 
my $pinyin = $opt_p;         # print pin-yin

$from_enc = 'gb2312-raw' if $from_enc =~ /big5/i;
my @supported = Encode->encodings("CN");
push @supported, qw(utf8 utf-8 utf-16);
my %supported = map { $_ => 1 } @supported;

die "Unknown 'from' encoding $from_enc\n" unless $supported{lc($from_enc)};
die "Unknown 'to'   encoding $to_enc\n"   unless $supported{lc($to_enc)};

my $dict = Lingua::ZH::CEDICT->new();
$dict->init();

# Globals
my $lnum = 0;
my $offset = 0;
my $count = 0;    # match count

sub blength {
    my $s = shift;
    {
	use bytes;
	return length($s);
    }
}

sub recover_bytes {
    my($raw, $dec) = @_;
    
    my @chars = (); # raw bytes
    my $ch;         # accumulate bytes until a valid character
    my $i = 0;      # index into raw string

    my $j = 0;      # index into decoded string
    my $dcheck;

    my $BADCHAR = decode('utf8', "\xff\xfd");

    while( $i < length($raw) ){
	$ch .= substr($raw, $i, 1);          # eat a byte off raw string
	my $dch = decode($from_enc, $ch);    # this may be invalid
	if( $j >= length($dec) ){
#	    print "Decoded string ran out at $j\n";
	    last;
	}
	my $dec_char = substr($dec, $j, 1);
	if( $dec_char eq $BADCHAR ){         # the following few bytes are invalid
	                                     #  skip them
	    
#	    HOW MANY BYTES TO SKIP?
#            until next prefix matches next decoded character. 
#	    are there ever several consequtive undefined characters in decoded line?
#            that would break it, but otherwise it should work
	}
	if( $dch eq $dec_char ){             # is it the same symbol as decoded?
	    $dcheck .= $dec_char;            # "check string"
	    push @chars, $ch;
	    $ch = '';
	    $j++;                            # consume decoded char
	}
	$i++;
    }

    unless( $dcheck eq $dec ){ # find the first index where they deviate
#	print "Lengths: check ", length($dcheck), " decoded ", length($dec), "\n";
	for($i=0; $i < length($dcheck) && $i < length($dec); $i++ ){
	    unless( substr($dcheck, $i, 1) eq substr($dec, $i, 1) ){
#		print substr($dcheck, $i, 1), ' VS ', substr($dec, $i, 1), "\n";
		last;
	    }
	}
#	print $i, ": ";
#	print u2h(substr($raw, 0)), " | ", join(' ', map { u2h($_) } @chars ), "\n\n";
#	
#	print u2h(decode($from_enc, substr($raw, $i))), "\n\n";
#	print u2h($dec), "\n\n";

    }
    return @chars;
}

#
#  Main loop
#
while(<>){
    $lnum++;
    my $rawline = $_;

    my $line = decode($from_enc, $_ );

    my @rawchars = recover_bytes($rawline, $line);

    my $i = 0;                        # position in string
    while( $i<length($line) ){
	my $c = substr($line, $i, 1); # bite off a character

	if( ord($c) < 128 ){ # no point in looking up ascii
	    $i++;
	    next;
	}

	## should skip all non-Chinese unicode 
	unless( ord($c) >= 0x4e00 && ord($c) <= 0x9fff ){
	    $i++;
	    next;
	}

	print "($lnum, $i)/", encode($to_enc, $c), " 0x", u2h($c), "\n" if $dbg;

	my @results = ();

	$dict->startMatch("^$c");

	while (my $e = $dict->match()) {

	    # see how far it matches into the actual string
	    my $m_t = decode('UTF-8', $e->[0]);  # traditional match
	    my $m_s = decode('UTF-8', $e->[1]);  # simplified  match
	    my $j = 0;

	    print $m_t, " " , substr($line, $i), " ",
	                      substr($m_t, $j), 
	                      " $i \n" if $dbg;

	    # the dictionary returns both a traditional and a simplified
	    #   form for each match. We must save the flag to reflect
	    #   which actually matched

	    while( $j < length($m_t) &&  # this assumes m_t and m_s are same size!
		   $i+$j < length($line) && 
		   ( substr($line, $i+$j, 1) eq substr($m_t, $j, 1) || 
		     substr($line, $i+$j, 1) eq substr($m_s, $j, 1) 
		   )){
		$e->[6] = 1 if substr($line, $i+$j, 1) eq substr($m_s, $j, 1);
		$j++;
	    }
	    $e->[5]=$j;  # save the matching length in an extra member
	    #   of the lookup result structure, for sorting later

	    print $m_t, " " , substr($line, $i+$j), " ",
	                      substr($m_t, $j), 
	                      " $i $j \n" if $dbg;
	    
	    push @results, $e;	    
	}

	unless( @results ){  # no results, skip this symbol
	    $i++; next;
	}
	
	# properly speaking, this should be a dynamic programming exercise,
	#   but we punt on that for now. Just pick the longest match
	#
	# Note: a match may contain chars not in the file! Fix that
	       
	my @sorted = sort { $b->[5] <=> $a->[5] } @results;
	if( $dbg ){
	    for my $e (@sorted){ 
		#  match_lenth    trad    simp    pinyin pinyin w/o tones  english
		print "$e->[5] : $e->[0] ". u2h($e->[0]) . " $e->[1] " . u2h($e->[1]) ." [$e->[2] / $e->[3]] $e->[4]\n";
	    }
	}

	my $f = $sorted[0];  # first match	

	warn "Zero-length match, something went wrong! See FIXMEs in the code :(\n" unless $f->[5] >= 1;
	
	### Now print the match, based on options

	my $m = ($f->[6]) ? $f->[1] : $f->[0]; # traditional or simple?

	# dirty hack, to find the _byte_ not _char_ position in input string
	#  is this only approximate, due to encoding differences
	my $str2ptr = substr($line, 0, $i);
	my $byte_off = length(encode($from_enc, $str2ptr));

	my @out; # line of output, be be join-ed
	push @out, $lnum                 if $line_number;
	push @out, $i                    if $column_number;
	push @out, ($offset + $byte_off) if $byte_offset;
	push @out, $m;                   # Chinese match                   
	push @out, $f->[2]               if $pinyin;

	#push @out, u2h(substr($rawline, 
	#		      $byte_off, 1)) if $hex;

	push @out, $f->[4];              # English

	print join(" ", @out), "\n";

	# print "At ($lnum, $i, ". ($offset+$i) . "): $f->[0] ". u2h($f->[0]) . " $f->[1] " . u2h($f->[1]) ." [$f->[2] / $f->[3]] $f->[4]\n" ;

	$i += $f->[5];  # advance the column number/string position

	if( $max_count && $count++ > $max_count ){
	    die "OMG, its full of APT! $max_count matches exceeded!\n";
	}
    }
    $offset += length($rawline);
}

sub u2h {
    my $str = shift;
    my $res;
    for(my $i=0; $i<length($str); $i++){
	my $c = substr($str, $i, 1);
	$res .= sprintf("%x ", ord($c));
    }
    return $res;
}
